@@ -240,16 +240,17 @@ pango_font_get_hb_font_for_context (PangoFont           *font,
 
   hb_font = pango_font_get_hb_font (font);
 
-  if (G_UNLIKELY (!funcs))
+  if (G_UNLIKELY (g_once_init_enter (&funcs)))
     {
-      funcs = hb_font_funcs_create ();
+      hb_font_funcs_t *f = hb_font_funcs_create ();
 
-      hb_font_funcs_set_nominal_glyph_func (funcs, pango_hb_font_get_nominal_glyph, NULL, NULL);
-      hb_font_funcs_set_glyph_h_advance_func (funcs, pango_hb_font_get_glyph_h_advance, NULL, NULL);
-      hb_font_funcs_set_glyph_v_advance_func (funcs, pango_hb_font_get_glyph_v_advance, NULL, NULL);
-      hb_font_funcs_set_glyph_extents_func (funcs, pango_hb_font_get_glyph_extents, NULL, NULL);
+      hb_font_funcs_set_nominal_glyph_func (f, pango_hb_font_get_nominal_glyph, NULL, NULL);
+      hb_font_funcs_set_glyph_h_advance_func (f, pango_hb_font_get_glyph_h_advance, NULL, NULL);
+      hb_font_funcs_set_glyph_v_advance_func (f, pango_hb_font_get_glyph_v_advance, NULL, NULL);
+      hb_font_funcs_set_glyph_extents_func (f, pango_hb_font_get_glyph_extents, NULL, NULL);
 
-      hb_font_funcs_make_immutable (funcs);
+      hb_font_funcs_make_immutable (f);
+      g_once_init_leave (&funcs, f);
     }
 
   context->font = font;
@@ -298,6 +299,23 @@ find_text_transform (const PangoAnalysis *analysis)
   return transform;
 }
 
+static gboolean
+font_has_color (hb_font_t *font)
+{
+  hb_face_t *face;
+
+  face = hb_font_get_face (font);
+
+#if HB_VERSION_ATLEAST (7, 0, 0)
+  if (hb_ot_color_has_paint (face))
+    return TRUE;
+#endif
+
+  return hb_ot_color_has_layers (face) ||
+         hb_ot_color_has_png (face) ||
+         hb_ot_color_has_svg (face);
+}
+
 static gboolean
 glyph_has_color (hb_font_t      *font,
                  hb_codepoint_t  glyph)
@@ -307,6 +325,11 @@ glyph_has_color (hb_font_t      *font,
 
   face = hb_font_get_face (font);
 
+#if HB_VERSION_ATLEAST (7, 0, 0)
+  if (hb_ot_color_glyph_has_paint (face, glyph))
+    return TRUE;
+#endif
+
   if (hb_ot_color_glyph_get_layers (face, glyph, 0, NULL, NULL) > 0)
     return TRUE;
 
@@ -366,6 +389,7 @@ pango_hb_shape (const char          *item_text,
   PangoGlyphInfo *infos;
   PangoTextTransform transform;
   int hyphen_index;
+  gboolean font_is_color;
 
   g_return_if_fail (analysis != NULL);
   g_return_if_fail (analysis->font != NULL);
@@ -403,7 +427,7 @@ pango_hb_shape (const char          *item_text,
 
       hyphen_index = item_offset + item_length - last_char_len;
 
-      if (log_attrs[num_chars].break_removes_preceding)
+      if (log_attrs && log_attrs[num_chars].break_removes_preceding)
         item_length -= last_char_len;
     }
 
@@ -443,7 +467,7 @@ pango_hb_shape (const char          *item_text,
               break;
 
             case PANGO_TEXT_TRANSFORM_CAPITALIZE:
-              if (log_attrs[i].is_word_start)
+              if (log_attrs && log_attrs[i].is_word_start)
                 ch = g_unichar_totitle (ch);
               break;
 
@@ -496,13 +520,14 @@ pango_hb_shape (const char          *item_text,
   pango_glyph_string_set_size (glyphs, num_glyphs);
   infos = glyphs->glyphs;
   last_cluster = -1;
+  font_is_color = font_has_color (hb_font);
 
   for (i = 0; i < num_glyphs; i++)
     {
       infos[i].glyph = hb_glyph->codepoint;
       glyphs->log_clusters[i] = hb_glyph->cluster - item_offset;
       infos[i].attr.is_cluster_start = glyphs->log_clusters[i] != last_cluster;
-      infos[i].attr.is_color = glyph_has_color (hb_font, hb_glyph->codepoint);
+      infos[i].attr.is_color = font_is_color && glyph_has_color (hb_font, hb_glyph->codepoint);
       hb_glyph++;
       last_cluster = glyphs->log_clusters[i];
     }
@@ -784,6 +809,10 @@ pango_shape_internal (const char          *item_text,
  * that API allows for shaping interaction happening across text item
  * boundaries.
  *
+ * Some aspects of hyphen insertion and text transformation (in particular,
+ * capitalization) require log attrs, and thus can only be handled by
+ * [func@Pango.shape_item].
+ *
  * Note that the extra attributes in the @analyis that is returned from
  * [func@Pango.itemize] have indices that are relative to the entire paragraph,
  * so you need to subtract the item offset from their indices before
@@ -819,6 +848,10 @@ pango_shape (const char          *text,
  * text of which @item_text is part of, provide the broader text as
  * @paragraph_text. If @paragraph_text is %NULL, item text is used instead.
  *
+ * Some aspects of hyphen insertion and text transformation (in particular,
+ * capitalization) require log attrs, and thus can only be handled by
+ * [func@Pango.shape_item].
+ *
  * Note that the extra attributes in the @analyis that is returned from
  * [func@Pango.itemize] have indices that are relative to the entire paragraph,
  * so you do not pass the full paragraph text as @paragraph_text, you need
@@ -863,6 +896,10 @@ pango_shape_full (const char          *item_text,
  * This is similar to [func@Pango.shape_full], except it also takes flags
  * that can influence the shaping process.
  *
+ * Some aspects of hyphen insertion and text transformation (in particular,
+ * capitalization) require log attrs, and thus can only be handled by
+ * [func@Pango.shape_item].
+ *
  * Note that the extra attributes in the @analyis that is returned from
  * [func@Pango.itemize] have indices that are relative to the entire paragraph,
  * so you do not pass the full paragraph text as @paragraph_text, you need
@@ -900,8 +937,9 @@ pango_shape_with_flags (const char          *item_text,
  *
  * This is similar to [func@Pango.shape_with_flags], except it takes a
  * `PangoItem` instead of separate @item_text and @analysis arguments.
- * It also takes @log_attrs, which may be used in implementing text
- * transforms.
+ *
+ * It also takes @log_attrs, which are needed for implementing some aspects
+ * of hyphen insertion and text transforms (in particular, capitalization).
  *
  * Note that the extra attributes in the @analyis that is returned from
  * [func@Pango.itemize] have indices that are relative to the entire paragraph,
